@@ -1,6 +1,6 @@
 /* dirmngr.c - Keyserver and X.509 LDAP access
  * Copyright (C) 2002 Klarälvdalens Datakonsult AB
- * Copyright (C) 2003-2004, 2006-2007, 2008, 2010-2011, 2020-2021 g10 Code GmbH
+ * Copyright (C) 2003-2004, 2006-2007, 2008, 2010-2011, 2020 g10 Code GmbH
  * Copyright (C) 2014 Werner Koch
  *
  * This file is part of GnuPG.
@@ -71,6 +71,7 @@
 #if USE_LDAP
 # include "ldap-wrapper.h"
 #endif
+#include "../common/comopt.h"
 #include "../common/init.h"
 #include "../common/gc-opt-flags.h"
 #include "dns-stuff.h"
@@ -158,13 +159,15 @@ enum cmd_and_opt_values {
   oConnectTimeout,
   oConnectQuickTimeout,
   oListenBacklog,
+  oFakeCRL,
   oCompatibilityFlags,
+  oUserAgent,
   aTest
 };
 
 
 
-static ARGPARSE_OPTS opts[] = {
+static gpgrt_opt_t opts[] = {
 
   ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
   ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
@@ -175,7 +178,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_c (aServer,   "server",  N_("run in server mode (foreground)") ),
   ARGPARSE_c (aDaemon,   "daemon",  N_("run in daemon mode (background)") ),
 #ifndef HAVE_W32_SYSTEM
-  ARGPARSE_c (aSupervised,  "supervised", N_("run in supervised mode")),
+  ARGPARSE_c (aSupervised,  "deprecated-supervised", "@"),
 #endif
   ARGPARSE_c (aListCRLs, "list-crls", N_("list the contents of the CRL cache")),
   ARGPARSE_c (aLoadCRL,  "load-crl", N_("|FILE|load CRL from FILE into cache")),
@@ -219,7 +222,7 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
   ARGPARSE_s_i (oMaxReplies, "max-replies",
                 N_("|N|do not return more than N items in one query")),
-  ARGPARSE_s_u (oFakedSystemTime, "faked-system-time", "@"), /*(epoch time)*/
+  ARGPARSE_s_s (oFakedSystemTime, "faked-system-time", "@"),
   ARGPARSE_s_n (oDisableCheckOwnSocket, "disable-check-own-socket", "@"),
   ARGPARSE_s_s (oIgnoreCert,"ignore-cert", "@"),
   ARGPARSE_s_s (oIgnoreCertExtension,"ignore-cert-extension", "@"),
@@ -240,14 +243,6 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oConnectQuickTimeout, "connect-quick-timeout", "@"),
 
 
-  ARGPARSE_header ("Keyserver", N_("Configuration for Keyservers")),
-
-  ARGPARSE_s_s (oKeyServer, "keyserver",
-                N_("|URL|use keyserver at URL")),
-  ARGPARSE_s_s (oHkpCaCert, "hkp-cacert",
-                N_("|FILE|use the CA certificates in FILE for HKP over TLS")),
-
-
   ARGPARSE_header ("HTTP", N_("Configuration for HTTP servers")),
 
   ARGPARSE_s_n (oDisableHTTP, "disable-http", N_("inhibit the use of HTTP")),
@@ -257,10 +252,17 @@ static ARGPARSE_OPTS opts[] = {
                 N_("|URL|redirect all HTTP requests to URL")),
   ARGPARSE_s_n (oHonorHTTPProxy, "honor-http-proxy",
                 N_("use system's HTTP proxy setting")),
+  ARGPARSE_s_s (oUserAgent, "user-agent", "@"),
   ARGPARSE_s_s (oLDAPWrapperProgram, "ldap-wrapper-program", "@"),
 
+  ARGPARSE_header ("Keyserver", N_("Configuration for OpenPGP servers")),
 
-  ARGPARSE_header ("LDAP", N_("Configuration of LDAP servers to use")),
+  ARGPARSE_s_s (oKeyServer, "keyserver",
+                N_("|URL|use keyserver at URL")),
+  ARGPARSE_s_s (oHkpCaCert, "hkp-cacert",
+                N_("|FILE|use the CA certificates in FILE for HKP over TLS")),
+
+  ARGPARSE_header ("LDAP", N_("Configuration for X.509 servers")),
 
   ARGPARSE_s_n (oDisableLDAP, "disable-ldap", N_("inhibit the use of LDAP")),
   ARGPARSE_s_n (oIgnoreLDAPDP,"ignore-ldap-dp",
@@ -278,7 +280,7 @@ static ARGPARSE_OPTS opts[] = {
                    " points to serverlist")),
   ARGPARSE_s_i (oLDAPTimeout, "ldaptimeout",
                 N_("|N|set LDAP timeout to N seconds")),
-
+  ARGPARSE_s_s (oFakeCRL, "fake-crl", "@"),
 
   ARGPARSE_header ("OCSP", N_("Configuration for OCSP")),
 
@@ -329,13 +331,15 @@ static struct debug_flags_s debug_flags [] =
     { DBG_NETWORK_VALUE, "network" },
     { DBG_LOOKUP_VALUE , "lookup"  },
     { DBG_EXTPROG_VALUE, "extprog" },
+    { DBG_KEEPTMP_VALUE, "keeptmp" },
+    { DBG_LDAP_VALUE,    "ldap"    },
     { 77, NULL } /* 77 := Do not exit on "help" or "?".  */
   };
 
 /* The list of compatibility flags.  */
 static struct compatibility_flags_s compatibility_flags [] =
   {
-    { COMPAT_OCSP_SHA256_CERTID, "ocsp-sha256-certid" },
+    { COMPAT_RESTRICT_HTTP_REDIR, "restrict-http-redir" },
     { 0, NULL }
   };
 
@@ -393,6 +397,9 @@ static enum
   } tor_mode;
 
 
+/* Flag indicating that we are in supervised mode.  */
+static int is_supervised;
+
 /* Counter for the active connections.  */
 static int active_connections;
 
@@ -400,7 +407,7 @@ static int active_connections;
  * thread to run background network tasks.  */
 static int network_activity_seen;
 
-/* A list of filenames registred with --hkp-cacert.  */
+/* A list of filenames registered with --hkp-cacert.  */
 static strlist_t hkp_cacert_filenames;
 
 /* A flag used to clear the list of ldapservers iff --ldapserver is
@@ -449,9 +456,6 @@ static void handle_connections (assuan_fd_t listen_fd);
 static void gpgconf_versions (void);
 
 
-/* NPth wrapper function definitions. */
-ASSUAN_SYSTEM_NPTH_IMPL;
-
 static const char *
 my_strusage( int level )
 {
@@ -483,9 +487,9 @@ my_strusage( int level )
 
 
 /* Callback from libksba to hash a provided buffer.  Our current
- * implementation does only allow SHA-1 and SHA-256 for hashing. This
- * may be extended by mapping the name, testing for algorithm
- * availibility and adjust the length checks accordingly.  */
+   implementation does only allow SHA-1 for hashing. This may be
+   extended by mapping the name, testing for algorithm availability
+   and adjust the length checks accordingly. */
 static gpg_error_t
 my_ksba_hash_buffer (void *arg, const char *oid,
                      const void *buffer, size_t length, size_t resultsize,
@@ -493,22 +497,12 @@ my_ksba_hash_buffer (void *arg, const char *oid,
 {
   (void)arg;
 
-  if (!oid || !strcmp (oid, "1.3.14.3.2.26"))
-    {
-      if (resultsize < 20)
-        return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
-      gcry_md_hash_buffer (GCRY_MD_SHA1, result, buffer, length);
-      *resultlen = 20;
-    }
-  else if (!strcmp (oid, "2.16.840.1.101.3.4.2.1"))
-    {
-      if (resultsize < 32)
-        return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
-      gcry_md_hash_buffer (GCRY_MD_SHA256, result, buffer, length);
-      *resultlen = 32;
-    }
-  else
+  if (oid && strcmp (oid, "1.3.14.3.2.26"))
     return gpg_error (GPG_ERR_NOT_SUPPORTED);
+  if (resultsize < 20)
+    return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
+  gcry_md_hash_buffer (2, result, buffer, length);
+  *resultlen = 20;
   return 0;
 }
 
@@ -557,7 +551,7 @@ set_debug (void)
          select the highest debug value and would then clutter their
          disk with debug files which may reveal confidential data.  */
       if (numok)
-        opt.debug &= ~(DBG_HASHING_VALUE);
+        opt.debug &= ~(DBG_HASHING_VALUE|DBG_KEEPTMP_VALUE);
     }
   else
     {
@@ -603,7 +597,7 @@ set_tor_mode (void)
 {
   if (dirmngr_use_tor ())
     {
-      /* Enable Tor mode and when called again force a new curcuit
+      /* Enable Tor mode and when called again force a new circuit
        * (e.g. on SIGHUP).  */
       enable_dns_tormode (1);
       if (assuan_sock_set_flag (ASSUAN_INVALID_FD, "tor-mode", 1))
@@ -680,7 +674,7 @@ shutdown_reaper (void)
    PARGS, resets the options to the default.  REREAD should be set
    true if it is not the initial option parsing. */
 static int
-parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
+parse_rereadable_options (gpgrt_argparse_t *pargs, int reread)
 {
   if (!pargs)
     { /* Reset mode. */
@@ -704,6 +698,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.ocsp_max_period = 90 * 86400;       /* 90 days.  */
       opt.ocsp_current_period = 3 * 60 * 60;  /* 3 hours. */
       opt.max_replies = DEFAULT_MAX_REPLIES;
+      opt.user_agent = "GnuPG/2.6";
       while (opt.ocsp_signer)
         {
           fingerprint_list_t tmp = opt.ocsp_signer->next;
@@ -732,6 +727,8 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
       ldapserver_list_needs_reset = 1;
       opt.debug_cache_expired_certs = 0;
+      xfree (opt.fake_crl);
+      opt.fake_crl = NULL;
       opt.compat_flags = 0;
       return 1;
     }
@@ -899,6 +896,11 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.debug_cache_expired_certs = 0;
       break;
 
+    case oFakeCRL:
+      xfree (opt.fake_crl);
+      opt.fake_crl = *pargs->r.ret_str? xstrdup (pargs->r.ret_str) : NULL;
+      break;
+
     case oCompatibilityFlags:
       if (parse_compatibility_flags (pargs->r.ret_str, &opt.compat_flags,
                                      compatibility_flags))
@@ -908,6 +910,15 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
         }
       break;
 
+    case oUserAgent:
+      if (strpbrk (pargs->r.ret_str, "\r\n"))
+        ; /* Ignore if the caller tried to insert CR or LF.  */
+      else if (!strcmp (pargs->r.ret_str, "none"))
+        opt.user_agent = "";
+      else
+        opt.user_agent = pargs->r.ret_str;
+      break;
+
     default:
       return 0; /* Not handled. */
     }
@@ -921,7 +932,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
 }
 
 
-/* This fucntion is called after option parsing to adjust some values
+/* This function is called after option parsing to adjust some values
  * and call option setup functions.  */
 static void
 post_option_parsing (enum cmd_and_opt_values cmd)
@@ -967,40 +978,64 @@ my_ntbtls_log_handler (void *opaque, int level, const char *fmt, va_list argv)
   (void)opaque;
 
   if (level == -1)
-    log_logv_with_prefix (GPGRT_LOG_INFO, "ntbtls: ", fmt, argv);
+    log_logv_prefix (GPGRT_LOGLVL_INFO, "ntbtls: ", fmt, argv);
   else
     {
       char prefix[10+20];
       snprintf (prefix, sizeof prefix, "ntbtls(%d): ", level);
-      log_logv_with_prefix (GPGRT_LOG_DEBUG, prefix, fmt, argv);
+      log_logv_prefix (GPGRT_LOGLVL_DEBUG, prefix, fmt, argv);
     }
 }
 #endif
 
 
+/* Helper for initialize_modules.  */
 static void
 thread_init (void)
 {
-  npth_init ();
-  assuan_set_system_hooks (ASSUAN_SYSTEM_NPTH);
+  static int npth_initialized = 0;
+
+  if (!npth_initialized)
+    {
+      npth_initialized++;
+      npth_init ();
+      /* With nPth running we can set the logging callback.  Our
+       * windows implementation does not yet feature the nPth TLS
+       * functions.  */
+#ifndef HAVE_W32_SYSTEM
+      if (npth_key_create (&my_tlskey_current_fd, NULL) == 0)
+        if (npth_setspecific (my_tlskey_current_fd, NULL) == 0)
+          log_set_pid_suffix_cb (pid_suffix_callback);
+#endif /*!HAVE_W32_SYSTEM*/
+      }
   gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
 
-  /* Now with NPth running we can set the logging callback.  Our
-     windows implementation does not yet feature the NPth TLS
-     functions.  */
-#ifndef HAVE_W32_SYSTEM
-  if (npth_key_create (&my_tlskey_current_fd, NULL) == 0)
-    if (npth_setspecific (my_tlskey_current_fd, NULL) == 0)
-      log_set_pid_suffix_cb (pid_suffix_callback);
-#endif /*!HAVE_W32_SYSTEM*/
+  /* Now that we have set the syscall clamp we need to tell Libgcrypt
+   * that it should get them from libgpg-error.  Note that Libgcrypt
+   * has already been initialized but at that point nPth was not
+   * initialized and thus Libgcrypt could not set its system call
+   * clamp.  */
+  gcry_control (GCRYCTL_REINIT_SYSCALL_CLAMP, 0, 0);
+  assuan_control (ASSUAN_CONTROL_REINIT_SYSCALL_CLAMP, NULL);
 }
 
 
+static void
+initialize_modules (void)
+{
+  thread_init ();
+  cert_cache_init (hkp_cacert_filenames);
+  crl_cache_init ();
+  ks_hkp_init ();
+}
+
+
+
 int
 main (int argc, char **argv)
 {
   enum cmd_and_opt_values cmd = 0;
-  ARGPARSE_ARGS pargs;
+  gpgrt_argparse_t pargs;
   int orig_argc;
   char **orig_argv;
   char *last_configname = NULL;
@@ -1020,7 +1055,7 @@ main (int argc, char **argv)
   struct assuan_malloc_hooks malloc_hooks;
 
   early_system_init ();
-  set_strusage (my_strusage);
+  gpgrt_set_strusage (my_strusage);
   log_set_prefix (DIRMNGR_NAME, GPGRT_LOG_WITH_PREFIX | GPGRT_LOG_WITH_PID);
 
   /* Make sure that our subsystems are ready.  */
@@ -1089,7 +1124,7 @@ main (int argc, char **argv)
   pargs.argc = &argc;
   pargs.argv = &argv;
   pargs.flags= (ARGPARSE_FLAG_KEEP | ARGPARSE_FLAG_NOVERSION);
-  while (gnupg_argparse (NULL, &pargs, opts))
+  while (gpgrt_argparse (NULL, &pargs, opts))
     {
       switch (pargs.r_opt)
         {
@@ -1107,12 +1142,12 @@ main (int argc, char **argv)
 
   socket_name = dirmngr_socket_name ();
 
-  /* The configuraton directories for use by gpgrt_argparser.  */
-  gnupg_set_confdir (GNUPG_CONFDIR_SYS, gnupg_sysconfdir ());
-  gnupg_set_confdir (GNUPG_CONFDIR_USER, gnupg_homedir ());
+  /* The configuration directories for use by gpgrt_argparser.  */
+  gpgrt_set_confdir (GPGRT_CONFDIR_SYS, gnupg_sysconfdir ());
+  gpgrt_set_confdir (GPGRT_CONFDIR_USER, gnupg_homedir ());
 
   /* We are re-using the struct, thus the reset flag.  We OR the
-   * flags so that the internal intialized flag won't be cleared. */
+   * flags so that the internal initialized flag won't be cleared. */
   argc = orig_argc;
   argv = orig_argv;
   pargs.argc = &argc;
@@ -1121,7 +1156,7 @@ main (int argc, char **argv)
                    | ARGPARSE_FLAG_KEEP
                    | ARGPARSE_FLAG_SYS
                    | ARGPARSE_FLAG_USER);
-  while (gnupg_argparser (&pargs, opts, DIRMNGR_NAME EXTSEP_S "conf"))
+  while (gpgrt_argparser (&pargs, opts, DIRMNGR_NAME EXTSEP_S "conf"))
     {
       if (pargs.r_opt == ARGPARSE_CONFFILE)
         {
@@ -1178,7 +1213,12 @@ main (int argc, char **argv)
 	case oLDAPAddServers: opt.add_new_ldapservers = 1; break;
 
         case oFakedSystemTime:
-          gnupg_set_time ((time_t)pargs.r.ret_ulong, 0);
+          {
+            time_t faked_time = isotime2epoch (pargs.r.ret_str);
+            if (faked_time == (time_t)(-1))
+              faked_time = (time_t)strtoul (pargs.r.ret_str, NULL, 10);
+            gnupg_set_time (faked_time, 0);
+          }
           break;
 
         case oForce: opt.force = 1; break;
@@ -1197,12 +1237,12 @@ main (int argc, char **argv)
           break;
 	}
     }
-  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
+  gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
 
   if (!last_configname)
-    opt.config_filename = make_filename (gnupg_homedir (),
-                                         DIRMNGR_NAME EXTSEP_S "conf",
-                                         NULL);
+    opt.config_filename = gpgrt_fnameconcat (gnupg_homedir (),
+                                             DIRMNGR_NAME EXTSEP_S "conf",
+                                             NULL);
   else
     {
       opt.config_filename = last_configname;
@@ -1211,6 +1251,14 @@ main (int argc, char **argv)
 
   if (log_get_errorcount(0))
     exit(2);
+
+  /* Get a default log file from common.conf.  */
+  if (!logfile && !parse_comopt (GNUPG_MODULE_NAME_DIRMNGR, debug_argparser))
+    {
+      logfile = comopt.logfile;
+      comopt.logfile = NULL;
+    }
+
   if (nogreeting )
     greeting = 0;
 
@@ -1220,8 +1268,8 @@ main (int argc, char **argv)
   if (greeting)
     {
       es_fprintf (es_stderr, "%s %s; %s\n",
-                  strusage(11), strusage(13), strusage(14) );
-      es_fprintf (es_stderr, "%s\n", strusage(15) );
+                  gpgrt_strusage(11), gpgrt_strusage(13), gpgrt_strusage(14));
+      es_fprintf (es_stderr, "%s\n", gpgrt_strusage(15));
     }
 
 #ifdef IS_DEVELOPMENT_VERSION
@@ -1307,23 +1355,27 @@ main (int argc, char **argv)
           log_debug ("... okay\n");
         }
 
-
-      thread_init ();
-      cert_cache_init (hkp_cacert_filenames);
-      crl_cache_init ();
+      initialize_modules ();
       http_register_netactivity_cb (netactivity_action);
+
       start_command_handler (ASSUAN_INVALID_FD, 0);
       shutdown_reaper ();
     }
 #ifndef HAVE_W32_SYSTEM
   else if (cmd == aSupervised)
     {
+      struct stat statbuf;
+
+      if (!opt.quiet)
+        log_info(_("WARNING: \"%s\" is a deprecated option\n"), "--supervised");
+
+      is_supervised = 1;
+
       /* In supervised mode, we expect file descriptor 3 to be an
          already opened, listening socket.
 
          We will also not detach from the controlling process or close
          stderr; the supervisor should handle all of that.  */
-      struct stat statbuf;
       if (fstat (3, &statbuf) == -1 && errno == EBADF)
         {
           log_error ("file descriptor 3 must be validin --supervised mode\n");
@@ -1343,9 +1395,7 @@ main (int argc, char **argv)
       else
         log_set_prefix (NULL, 0);
 
-      thread_init ();
-      cert_cache_init (hkp_cacert_filenames);
-      crl_cache_init ();
+      initialize_modules ();
       http_register_netactivity_cb (netactivity_action);
       handle_connections (3);
       shutdown_reaper ();
@@ -1441,6 +1491,7 @@ main (int argc, char **argv)
 	log_error (_("error getting nonce for the socket\n"));
       if (rc == -1)
         {
+          log_libassuan_system_error (fd);
           log_error (_("error binding socket to '%s': %s\n"),
                      serv_addr.sun_path,
                      gpg_strerror (gpg_error_from_syserror ()));
@@ -1568,10 +1619,9 @@ main (int argc, char **argv)
             }
         }
 
-      thread_init ();
-      cert_cache_init (hkp_cacert_filenames);
-      crl_cache_init ();
+      initialize_modules ();
       http_register_netactivity_cb (netactivity_action);
+
       handle_connections (fd);
       shutdown_reaper ();
     }
@@ -1590,9 +1640,8 @@ main (int argc, char **argv)
       memset (&ctrlbuf, 0, sizeof ctrlbuf);
       dirmngr_init_default_ctrl (&ctrlbuf);
 
-      thread_init ();
-      cert_cache_init (hkp_cacert_filenames);
-      crl_cache_init ();
+      initialize_modules ();
+
       if (!argc)
         rc = crl_cache_load (&ctrlbuf, NULL);
       else
@@ -1613,9 +1662,8 @@ main (int argc, char **argv)
       memset (&ctrlbuf, 0, sizeof ctrlbuf);
       dirmngr_init_default_ctrl (&ctrlbuf);
 
-      thread_init ();
-      cert_cache_init (hkp_cacert_filenames);
-      crl_cache_init ();
+      initialize_modules ();
+
       rc = crl_fetch (&ctrlbuf, argv[0], &reader);
       if (rc)
         log_error (_("fetching CRL from '%s' failed: %s\n"),
@@ -1725,9 +1773,9 @@ dirmngr_deinit_default_ctrl (ctrl_t ctrl)
 /* Create a list of LDAP servers from the file FILENAME. Returns the
    list or NULL in case of errors.
 
-   The format fo such a file is line oriented where empty lines and
+   The format of such a file is line oriented where empty lines and
    lines starting with a hash mark are ignored.  All other lines are
-   assumed to be colon seprated with these fields:
+   assumed to be colon separated with these fields:
 
    1. field: Hostname
    2. field: Portnumber
@@ -1750,14 +1798,10 @@ parse_ldapserver_file (const char* filename, int ignore_enoent)
   fp = es_fopen (filename, "r");
   if (!fp)
     {
-      if (errno == ENOENT)
-        {
-          if (!ignore_enoent)
-            log_info ("No ldapserver file at: '%s'\n", filename);
-        }
+      if (ignore_enoent && gpg_err_code_from_syserror () == GPG_ERR_ENOENT)
+        ;
       else
-        log_error (_("error opening '%s': %s\n"), filename,
-                   strerror (errno));
+        log_info ("failed to open '%s': %s\n", filename, strerror (errno));
       return NULL;
     }
 
@@ -1955,17 +1999,18 @@ parse_fingerprint_item (const char *string,
    Fixme: Due to the way the argument parsing works, we create a
    memory leak here for all string type arguments.  There is currently
    no clean way to tell whether the memory for the argument has been
-   allocated or points into the process' original arguments.  Unless
+   allocated or points into the process's original arguments.  Unless
    we have a mechanism to tell this, we need to live on with this. */
 static void
 reread_configuration (void)
 {
-  ARGPARSE_ARGS pargs;
+  gpgrt_argparse_t pargs;
   char *twopart;
   int dummy;
+  int logfile_seen = 0;
 
   if (!opt.config_filename)
-    return; /* No config file. */
+    goto finish; /* No config file. */
 
   twopart = strconcat (DIRMNGR_NAME EXTSEP_S "conf" PATHSEP_S,
                        opt.config_filename, NULL);
@@ -1980,7 +2025,7 @@ reread_configuration (void)
   pargs.flags = (ARGPARSE_FLAG_KEEP
                  |ARGPARSE_FLAG_SYS
                  |ARGPARSE_FLAG_USER);
-  while (gnupg_argparser (&pargs, opts, twopart))
+  while (gpgrt_argparser (&pargs, opts, twopart))
     {
       if (pargs.r_opt == ARGPARSE_CONFFILE)
         {
@@ -1990,11 +2035,28 @@ reread_configuration (void)
       else if (pargs.r_opt < -1)
         pargs.err = ARGPARSE_PRINT_WARNING;
       else /* Try to parse this option - ignore unchangeable ones. */
-        parse_rereadable_options (&pargs, 1);
+        {
+          if (pargs.r_opt == oLogFile)
+            logfile_seen = 1;
+          parse_rereadable_options (&pargs, 1);
+        }
     }
-  gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
+  gpgrt_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
   xfree (twopart);
   post_option_parsing (0);
+
+ finish:
+  /* Get a default log file from common.conf.  */
+  if (!logfile_seen && !parse_comopt (GNUPG_MODULE_NAME_DIRMNGR, !!opt.debug))
+    {
+      if (!current_logfile || !comopt.logfile
+          || strcmp (current_logfile, comopt.logfile))
+        {
+          log_set_file (comopt.logfile);
+          xfree (current_logfile);
+          current_logfile = comopt.logfile? xtrystrdup (comopt.logfile) : NULL;
+        }
+    }
 }
 
 
@@ -2006,6 +2068,7 @@ dirmngr_sighup_action (void)
   log_info (_("SIGHUP received - "
               "re-reading configuration and flushing caches\n"));
   reread_configuration ();
+  set_tor_mode ();
   cert_cache_deinit (0);
   crl_cache_deinit ();
   cert_cache_init (hkp_cacert_filenames);
@@ -2039,8 +2102,9 @@ handle_signal (int signo)
       break;
 
     case SIGUSR1:
-      cert_cache_print_stats ();
-      domaininfo_print_stats ();
+      /* See also cmd_getinfo:"stats".  */
+      cert_cache_print_stats (NULL);
+      domaininfo_print_stats (NULL);
       break;
 
     case SIGUSR2:
@@ -2057,7 +2121,7 @@ handle_signal (int signo)
       if (shutdown_pending > 2)
         {
           log_info (_("shutdown forced\n"));
-          log_info ("%s %s stopped\n", strusage(11), strusage(13) );
+          log_info ("%s %s stopped\n", gpgrt_strusage(11), gpgrt_strusage(13));
           cleanup ();
           dirmngr_exit (0);
 	}
@@ -2065,7 +2129,7 @@ handle_signal (int signo)
 
     case SIGINT:
       log_info (_("SIGINT received - immediate shutdown\n"));
-      log_info( "%s %s stopped\n", strusage(11), strusage(13));
+      log_info( "%s %s stopped\n", gpgrt_strusage(11), gpgrt_strusage(13));
       cleanup ();
       dirmngr_exit (0);
       break;
@@ -2122,9 +2186,14 @@ housekeeping_thread (void *arg)
 }
 
 
-#if GPGRT_GCC_HAVE_PUSH_PRAGMA
+/* We try to enable correct overflow handling for signed int (commonly
+ * used for time_t). With gcc 4.2 -fno-strict-overflow was introduced
+ * and used here as a pragma.  Later gcc versions (gcc 6?) removed
+ * this as a pragma and -fwrapv was then suggested as a replacement
+ * for -fno-strict-overflow.  */
+#if GPGRT_HAVE_PRAGMA_GCC_PUSH
 # pragma GCC push_options
-# pragma GCC optimize ("no-strict-overflow")
+# pragma GCC optimize ("wrapv")
 #endif
 static int
 time_for_housekeeping_p (time_t curtime)
@@ -2142,7 +2211,7 @@ time_for_housekeeping_p (time_t curtime)
     }
   return 0;
 }
-#if GPGRT_GCC_HAVE_PUSH_PRAGMA
+#if GPGRT_HAVE_PRAGMA_GCC_PUSH
 # pragma GCC pop_options
 #endif
 
@@ -2192,7 +2261,7 @@ check_nonce (assuan_fd_t fd, assuan_sock_nonce_t *nonce)
   if (assuan_sock_check_nonce (fd, nonce))
     {
       log_info (_("error reading nonce on fd %d: %s\n"),
-                FD2INT (fd), strerror (errno));
+                FD_DBG (fd), strerror (errno));
       assuan_sock_close (fd);
       return -1;
     }
@@ -2226,7 +2295,7 @@ start_connection_thread (void *arg)
 
   active_connections++;
   if (opt.verbose)
-    log_info (_("handler for fd %d started\n"), FD2INT (fd));
+    log_info (_("handler for fd %d started\n"), FD_DBG (fd));
 
   session_id = ++last_session_id;
   if (!session_id)
@@ -2234,7 +2303,7 @@ start_connection_thread (void *arg)
   start_command_handler (fd, session_id);
 
   if (opt.verbose)
-    log_info (_("handler for fd %d terminated\n"), FD2INT (fd));
+    log_info (_("handler for fd %d terminated\n"), FD_DBG (fd));
   active_connections--;
 
   workqueue_run_post_session_tasks (session_id);
@@ -2337,7 +2406,7 @@ handle_connections (assuan_fd_t listen_fd)
      to full second.  */
   FD_ZERO (&fdset);
   FD_SET (FD2INT (listen_fd), &fdset);
-  nfd = FD2INT (listen_fd);
+  nfd = FD2NUM (listen_fd);
   if (my_inotify_fd != -1)
     {
       FD_SET (my_inotify_fd, &fdset);
@@ -2354,7 +2423,7 @@ handle_connections (assuan_fd_t listen_fd)
       /* Shutdown test.  */
       if (shutdown_pending)
         {
-          if (!active_connections)
+          if (!active_connections || is_supervised)
             break; /* ready */
 
           /* Do not accept new connections but keep on running the
@@ -2407,7 +2476,7 @@ handle_connections (assuan_fd_t listen_fd)
 	{
           log_error (_("npth_pselect failed: %s - waiting 1s\n"),
                      strerror (saved_errno));
-          npth_sleep (1);
+          gnupg_sleep (1);
           continue;
 	}
 
@@ -2438,11 +2507,13 @@ handle_connections (assuan_fd_t listen_fd)
           gnupg_fd_t fd;
 
           plen = sizeof paddr;
-	  fd = INT2FD (npth_accept (FD2INT(listen_fd),
-				    (struct sockaddr *)&paddr, &plen));
+	  fd = assuan_sock_accept (listen_fd,
+                                   (struct sockaddr *)&paddr, &plen);
 	  if (fd == GNUPG_INVALID_FD)
 	    {
-	      log_error ("accept failed: %s\n", strerror (errno));
+              gpg_error_t myerr = gpg_error_from_syserror ();
+              log_libassuan_system_error (listen_fd);
+	      log_error ("accept failed: %s\n", gpg_strerror (myerr));
 	    }
           else
             {
@@ -2453,7 +2524,7 @@ handle_connections (assuan_fd_t listen_fd)
               memset (&argval, 0, sizeof argval);
               argval.afd = fd;
               snprintf (threadname, sizeof threadname,
-                        "conn fd=%d", FD2INT(fd));
+                        "conn fd=%d", FD_DBG (fd));
 
               ret = npth_create (&thread, &tattr,
                                  start_connection_thread, argval.aptr);
@@ -2476,7 +2547,7 @@ handle_connections (assuan_fd_t listen_fd)
   if (listen_fd != GNUPG_INVALID_FD)
     assuan_sock_close (listen_fd);
   cleanup ();
-  log_info ("%s %s stopped\n", strusage(11), strusage(13));
+  log_info ("%s %s stopped\n", gpgrt_strusage(11), gpgrt_strusage(13));
 }
 
 const char*
