@@ -252,6 +252,14 @@ keyring_new (void *token)
   return hd;
 }
 
+void
+keyring_fp_close (KEYRING_HANDLE hd)
+{
+  if (!hd)
+    return;
+  iobuf_close (hd->current.iobuf);
+}
+
 void
 keyring_release (KEYRING_HANDLE hd)
 {
@@ -261,7 +269,6 @@ keyring_release (KEYRING_HANDLE hd)
     active_handles--;
     xfree (hd->word_match.name);
     xfree (hd->word_match.pattern);
-    iobuf_close (hd->current.iobuf);
     xfree (hd);
 }
 
@@ -598,7 +605,7 @@ keyring_insert_keyblock (KEYRING_HANDLE hd, KBNODE kb)
 
     /* Close this one otherwise we will lose the position for
      * a next search.  Fixme: it would be better to adjust the position
-     * after the write opertions.
+     * after the write operations.
      */
     iobuf_close (hd->current.iobuf);
     hd->current.iobuf = NULL;
@@ -638,7 +645,7 @@ keyring_delete_keyblock (KEYRING_HANDLE hd)
 
     /* close this one otherwise we will lose the position for
      * a next search.  Fixme: it would be better to adjust the position
-     * after the write opertions.
+     * after the write operations.
      */
     iobuf_close (hd->current.iobuf);
     hd->current.iobuf = NULL;
@@ -691,14 +698,14 @@ prepare_search (KEYRING_HANDLE hd)
            we can easily use search_next.  */
         if (gpg_err_code (hd->current.error) == GPG_ERR_LEGACY_KEY)
           {
-            if (DBG_LOOKUP)
+            if (DBG_KEYDB)
               log_debug ("%s: last error was GPG_ERR_LEGACY_KEY, clearing\n",
                          __func__);
             hd->current.error = 0;
           }
         else
           {
-            if (DBG_LOOKUP)
+            if (DBG_KEYDB)
               log_debug ("%s: returning last error: %s\n",
                          __func__, gpg_strerror (hd->current.error));
             return hd->current.error; /* still in error state */
@@ -708,7 +715,7 @@ prepare_search (KEYRING_HANDLE hd)
     if (hd->current.kr && !hd->current.eof) {
         if ( !hd->current.iobuf )
           {
-            if (DBG_LOOKUP)
+            if (DBG_KEYDB)
               log_debug ("%s: missing iobuf!\n", __func__);
             return GPG_ERR_GENERAL; /* Position invalid after a modify.  */
           }
@@ -717,7 +724,7 @@ prepare_search (KEYRING_HANDLE hd)
 
     if (!hd->current.kr && hd->current.eof)
       {
-        if (DBG_LOOKUP)
+        if (DBG_KEYDB)
           log_debug ("%s: EOF!\n", __func__);
         return -1; /* still EOF */
       }
@@ -725,7 +732,7 @@ prepare_search (KEYRING_HANDLE hd)
     if (!hd->current.kr) { /* start search with first keyring */
         hd->current.kr = hd->resource;
         if (!hd->current.kr) {
-          if (DBG_LOOKUP)
+          if (DBG_KEYDB)
             log_debug ("%s: keyring not available!\n", __func__);
           hd->current.eof = 1;
           return -1; /* keyring not available */
@@ -733,7 +740,7 @@ prepare_search (KEYRING_HANDLE hd)
         log_assert (!hd->current.iobuf);
     }
     else { /* EOF */
-        if (DBG_LOOKUP)
+        if (DBG_KEYDB)
           log_debug ("%s: EOF\n", __func__);
         iobuf_close (hd->current.iobuf);
         hd->current.iobuf = NULL;
@@ -968,7 +975,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
   int save_mode;
   off_t offset, main_offset;
   size_t n;
-  int need_uid, need_words, need_keyid, need_fpr, any_skip;
+  int need_uid, need_words, need_keyid, need_fpr, any_skip, need_grip;
   int pk_no, uid_no;
   int initial_skip;
   int scanned_from_start;
@@ -976,9 +983,10 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
   PKT_user_id *uid = NULL;
   PKT_public_key *pk = NULL;
   u32 aki[2];
+  unsigned char grip[KEYGRIP_LEN];
 
   /* figure out what information we need */
-  need_uid = need_words = need_keyid = need_fpr = any_skip = 0;
+  need_uid = need_words = need_keyid = need_fpr = any_skip = need_grip = 0;
   for (n=0; n < ndesc; n++)
     {
       switch (desc[n].mode)
@@ -998,8 +1006,6 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
         case KEYDB_SEARCH_MODE_LONG_KID:
           need_keyid = 1;
           break;
-        case KEYDB_SEARCH_MODE_FPR16:
-        case KEYDB_SEARCH_MODE_FPR20:
         case KEYDB_SEARCH_MODE_FPR:
           need_fpr = 1;
           break;
@@ -1007,8 +1013,11 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
           /* always restart the search in this mode */
           keyring_search_reset (hd);
           break;
+        case KEYDB_SEARCH_MODE_KEYGRIP:
+          need_grip = 1;
+          break;
         default: break;
-	}
+        }
       if (desc[n].skipfnc)
         {
           any_skip = 1;
@@ -1016,14 +1025,14 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
         }
     }
 
-  if (DBG_LOOKUP)
+  if (DBG_KEYDB)
     log_debug ("%s: need_uid = %d; need_words = %d; need_keyid = %d; need_fpr = %d; any_skip = %d\n",
                __func__, need_uid, need_words, need_keyid, need_fpr, any_skip);
 
   rc = prepare_search (hd);
   if (rc)
     {
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: prepare_search failed: %s (%d)\n",
                    __func__, gpg_strerror (rc), gpg_err_code (rc));
       return rc;
@@ -1032,12 +1041,12 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
   use_key_present_hash = !!key_present_hash;
   if (!use_key_present_hash)
     {
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: no offset table.\n", __func__);
     }
   else if (!key_present_hash_ready)
     {
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: initializing offset table. (need_keyid: %d => 1)\n",
                    __func__, need_keyid);
       need_keyid = 1;
@@ -1046,13 +1055,13 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
     {
       struct key_present *oi;
 
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: look up by long key id, checking cache\n", __func__);
 
       oi = key_present_hash_lookup (key_present_hash, desc[0].u.kid);
       if (!oi)
         { /* We know that we don't have this key */
-          if (DBG_LOOKUP)
+          if (DBG_KEYDB)
             log_debug ("%s: cache says not present\n", __func__);
           hd->found.kr = NULL;
           hd->current.eof = 1;
@@ -1097,7 +1106,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
   pk_no = uid_no = 0;
   initial_skip = 1; /* skip until we see the start of a keyblock */
   scanned_from_start = iobuf_tell (hd->current.iobuf) == 0;
-  if (DBG_LOOKUP)
+  if (DBG_KEYDB)
     log_debug ("%s: %ssearching from start of resource.\n",
                __func__, scanned_from_start ? "" : "not ");
   init_parse_packet (&parsectx, hd->current.iobuf);
@@ -1137,13 +1146,16 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
           pk = pkt.pkt.public_key;
           ++pk_no;
 
-          if (need_fpr) {
-            fingerprint_from_pk (pk, afp, &an);
-            while (an < 20) /* fill up to 20 bytes */
-              afp[an++] = 0;
-          }
+          if (need_fpr)
+            {
+              fingerprint_from_pk (pk, afp, &an);
+              while (an < 32) /* fill up to 32 bytes */
+                afp[an++] = 0;
+            }
           if (need_keyid)
             keyid_from_pk (pk, aki);
+          if (need_grip)
+            keygrip_from_pk (pk, grip, 0);
 
           if (use_key_present_hash
               && !key_present_hash_ready
@@ -1175,21 +1187,19 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
             break;
 
           case KEYDB_SEARCH_MODE_SHORT_KID:
-            if (pk && desc[n].u.kid[1] == aki[1])
-              goto found;
+            if (pk
+               && ((pk->fprlen == 32 && desc[n].u.kid[1] == aki[0])
+                   || (pk->fprlen != 32 && desc[n].u.kid[1] == aki[1])))
+                goto found;
             break;
           case KEYDB_SEARCH_MODE_LONG_KID:
             if (pk && desc[n].u.kid[0] == aki[0]
                 && desc[n].u.kid[1] == aki[1])
               goto found;
             break;
-          case KEYDB_SEARCH_MODE_FPR16:
-            if (pk && !memcmp (desc[n].u.fpr, afp, 16))
-              goto found;
-            break;
-          case KEYDB_SEARCH_MODE_FPR20:
           case KEYDB_SEARCH_MODE_FPR:
-            if (pk && !memcmp (desc[n].u.fpr, afp, 20))
+            if (pk && desc[n].fprlen >= 16 && desc[n].fprlen <= 32
+                && !memcmp (desc[n].u.fpr, afp, desc[n].fprlen))
               goto found;
             break;
           case KEYDB_SEARCH_MODE_FIRST:
@@ -1200,6 +1210,10 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
             if (pk)
               goto found;
             break;
+          case KEYDB_SEARCH_MODE_KEYGRIP:
+            if (pk && !memcmp (desc[n].u.grip, grip, KEYGRIP_LEN))
+              goto found;
+            break;
           default:
             rc = GPG_ERR_INV_ARG;
             goto found;
@@ -1211,7 +1225,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
       if (rc)
         goto real_found;
 
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: packet starting at offset %lld matched descriptor %zu\n"
                    , __func__, (long long)offset, n);
 
@@ -1224,7 +1238,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
           if (desc[n].skipfnc
               && desc[n].skipfnc (desc[n].skipfncvalue, aki, uid_no))
             {
-              if (DBG_LOOKUP)
+              if (DBG_KEYDB)
                 log_debug ("%s: skipping match: desc %zd's skip function returned TRUE\n",
                            __func__, n);
               break;
@@ -1237,7 +1251,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
  real_found:
   if (!rc)
     {
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: returning success\n", __func__);
       hd->found.offset = main_offset;
       hd->found.kr = hd->current.kr;
@@ -1246,7 +1260,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
     }
   else if (rc == -1)
     {
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: no matches (EOF)\n", __func__);
 
       hd->current.eof = 1;
@@ -1280,7 +1294,7 @@ keyring_search (KEYRING_HANDLE hd, KEYDB_SEARCH_DESC *desc,
     }
   else
     {
-      if (DBG_LOOKUP)
+      if (DBG_KEYDB)
         log_debug ("%s: error encountered during search: %s (%d)\n",
                    __func__, gpg_strerror (rc), rc);
       hd->current.error = rc;
@@ -1479,8 +1493,6 @@ keyring_rebuild_cache (ctrl_t ctrl, void *token, int noisy)
         {
           if (gpg_err_code (rc) == GPG_ERR_LEGACY_KEY)
             continue;  /* Skip legacy keys.  */
-          if (gpg_err_code (rc) == GPG_ERR_UNKNOWN_VERSION)
-            continue;  /* Skip keys with unknown version.  */
           log_error ("keyring_get_keyblock failed: %s\n", gpg_strerror (rc));
           goto leave;
         }
@@ -1589,6 +1601,7 @@ keyring_rebuild_cache (ctrl_t ctrl, void *token, int noisy)
   xfree (tmpfilename);
   xfree (bakfilename);
   release_kbnode (keyblock);
+  keyring_fp_close (hd);
   keyring_lock (hd, 0);
   keyring_release (hd);
   return rc;
