@@ -28,13 +28,12 @@ extern "C" {
 
 #include "../common/iobuf.h"
 #include "keybox-search-desc.h"
+#include "keybox-fwddecl.h"
 
 #ifdef KEYBOX_WITH_X509
 # include <ksba.h>
 #endif
 
-typedef struct keybox_handle *KEYBOX_HANDLE;
-
 
 typedef enum
   {
@@ -71,8 +70,7 @@ int keybox_is_writable (void *token);
 
 KEYBOX_HANDLE keybox_new_openpgp (void *token, int secret);
 KEYBOX_HANDLE keybox_new_x509 (void *token, int secret);
-void keybox_close_all_files (void *token);
-
+void keybox_fp_close (KEYBOX_HANDLE hd);
 void keybox_release (KEYBOX_HANDLE hd);
 void keybox_push_found_state (KEYBOX_HANDLE hd);
 void keybox_pop_found_state (KEYBOX_HANDLE hd);
@@ -84,11 +82,15 @@ gpg_error_t keybox_lock (KEYBOX_HANDLE hd, int yes, long timeout);
 /*-- keybox-file.c --*/
 /* Fixme: This function does not belong here: Provide a better
    interface to create a new keybox file.  */
-int _keybox_write_header_blob (estream_t fp, int openpgp_flag);
+gpg_error_t _keybox_write_header_blob (estream_t fp, int openpgp_flag);
 
 /*-- keybox-search.c --*/
+gpg_error_t keybox_get_data (KEYBOX_HANDLE hd,
+                             void **r_buffer, size_t *r_length,
+                             enum pubkey_types *r_pubkey_type,
+                             unsigned char *r_ubid);
 gpg_error_t keybox_get_keyblock (KEYBOX_HANDLE hd, iobuf_t *r_iobuf,
-                                 int *r_uid_no, int *r_pk_no);
+                                 int *r_pk_no, int *r_uid_no);
 #ifdef KEYBOX_WITH_X509
 int keybox_get_cert (KEYBOX_HANDLE hd, ksba_cert_t *ret_cert);
 #endif /*KEYBOX_WITH_X509*/
@@ -112,11 +114,13 @@ gpg_error_t keybox_update_keyblock (KEYBOX_HANDLE hd,
 #ifdef KEYBOX_WITH_X509
 int keybox_insert_cert (KEYBOX_HANDLE hd, ksba_cert_t cert,
                         unsigned char *sha1_digest);
+int keybox_update_cert (KEYBOX_HANDLE hd, ksba_cert_t cert,
+                        unsigned char *sha1_digest);
 #endif /*KEYBOX_WITH_X509*/
 int keybox_set_flags (KEYBOX_HANDLE hd, int what, int idx, unsigned int value);
 
 int keybox_delete (KEYBOX_HANDLE hd);
-int keybox_compress (KEYBOX_HANDLE hd);
+void keybox_compress_when_no_other_users (void *token, int for_openpgp);
 
 
 /*--  --*/
